#include <math.h>

/***********************************************************/
// Cox Ross Rubinstein
double __stdcall
CRR (const int OptionType,
  const double spot,
  const double strike,
  const double t,
  const double r,
  const double q,
  const double sigma,
  const int nSteps,
  double *vvec)
{
  double delt;
  double erdt;
  double ermqdt;
  double u;
  double d;
  double p;
  double pstar;
  double inv_erdt;
  double tmp;
  double uSquare;
  double uPower;
  double uPowerJ;
  int i;
  int j;
  double X;
  if (OptionType == 1 || OptionType == -1)
  {
  }
  else
    return(-0.1e1);
  if (0 < nSteps && 0.0e0 < spot && 0.0e0 < strike && 0.0e0 < t && 0.0e0 < sigma)
  {
  }
  else
    return(-0.2e1);
  X = (double)(strike / spot);
  delt = t / (double) nSteps;
  erdt = exp(r * delt);
  inv_erdt = 0.1e1 / erdt;
  ermqdt = exp((r - q) * delt);
  u = exp(sigma * sqrt(delt));
  d = 0.1e1 / u;
  p = (ermqdt - d) / (u - d);
  pstar = 0.1e1 - p;
  uSquare = u * u;
  uPower = pow(u, (double) (-2 - nSteps));
  for (i = 1; i <= nSteps + 1; i++)
  {
    uPower = (double)(uPower * uSquare);
    tmp = (double)((double) OptionType * (uPower - X));
    if (0.0e0 <= tmp)
      vvec[i - 1] = tmp;
    else
      vvec[i - 1] = 0.0e0;
  }
  uPowerJ = pow(u, (double) (-2 - nSteps));
  for (j = nSteps; j >= 1; j--)
  {
    uPowerJ = (double)(uPowerJ * u);
    uPower = (double)(uPowerJ);
    for (i = 1; i <= j; i++)
    {
      vvec[i - 1] = (double)((p * vvec[i] + pstar * vvec[i - 1]) * inv_erdt);
      uPower = (double)(uPower * uSquare);
      tmp = (double)((double) OptionType * (uPower - X));
      if (vvec[i - 1] <= tmp)
        vvec[i - 1] = tmp;
      else
      {
      }
    }
  }
  return(spot * vvec[0]);
}

/***********************************************************/
double __stdcall
BS_dOne(
	const double S,
	const double K,
	const double t,
	const double r,
	const double q,
	const double v)
{
	double sigma;

	sigma = v*sqrt(t);
	return (log(S/K) + (r-q)*t)/sigma + 0.5*sigma;
}

double __stdcall
BS_dTwo(
	const double S,
	const double K,
	const double t,
	const double r,
	const double q,
	const double v)
{
	return BS_dOne(S,K,t,r,q,v) - v*sqrt(t);
}

/***********************************************************/
// Leisen Reimer with PP normal Inversion
double __stdcall
Peizer_Pratt_Inversion(const double z, const int n)
{
	int nOdd;

	//nOdd = 2*floor(n/2)+1;
	nOdd = n;

	return 0.5e0 + (double) (z>0 ? 1 : (z<0 ? -1 : 0)) * sqrt(0.25e0 - 0.25e0
		* exp(-z * z * pow((double) n + 0.1e1 / 0.3e1 + 0.1e0 / (double) (n + 1), -0.2e1)
		* ((double) n + 0.1e1 / 0.6e1)));
}


double __stdcall
LR (const int OptionType,
  const double spot,
  const double strike,
  const double t,
  const double r,
  const double q,
  const double sigma,
  const int nSteps,
  double *vv)
{
	double delt, erdt, inv_erdt, ermqdt, u, d, d1, d2, p, pDash, pStar;
	double X, tmp, LR_Term, LR_Term_j, LR_Factor, LR_Factor_j;
	int n, i, j;

	n = 2*floor(nSteps/2)+1;

	X = strike / spot;

	delt = t / n;
	erdt = exp(r * delt);
	inv_erdt = 1 / erdt;
	ermqdt = exp((r - q) * delt);

	// parameters for Leisen-Reimer
	d2 = BS_dTwo(1, X, t, r, q, sigma);
	d1 = BS_dOne(1, X, t, r, q, sigma);
	p = Peizer_Pratt_Inversion(d2, n);
	pStar = 1 - p;
	pDash = Peizer_Pratt_Inversion(d1, n);
	u = ermqdt * pDash / p;
	d = (ermqdt - p * u) / (1 - p);

	LR_Factor = (pDash / (-1 + pDash) * (-1 + p) / p);
	LR_Factor_j = (-1 + p) / (-1 + pDash) / ermqdt;

	LR_Term = pow(ermqdt * (-1 + pDash) / (-1 + p), n);
	for ( i = 0; i<= n; i++)
		{
		tmp = OptionType * (LR_Term - X);
		// pay off
		if (0 <= tmp)
			vv[i] = tmp;
		else
			vv[i] = 0.0;
	  LR_Term = LR_Factor * LR_Term;
		}

	LR_Term_j = pow(1 / LR_Factor_j, n-1);
	for ( j = n - 1; 0 <= j; j--)
		{

		LR_Term = LR_Term_j;
		for (i = 0; i<=j; i++)
			{
		  vv[i] = (p * vv[i + 1] + pStar * vv[i]) * inv_erdt;
			tmp = OptionType * (LR_Term - X);
			// pay off
			if (vv[i] <= tmp)
				vv[i] = tmp;

			LR_Term = LR_Factor * LR_Term;
			}

		LR_Term_j = LR_Factor_j * LR_Term_j;
		}

	return spot*vv[0];
}
